(** * Table of Contents file *)
(** This file describes a correspondence between code and paper. *)
(** To build all of this code, you must have installed:

    - ocaml

    - Coq 8.7

    - Ltac2 plugin (available on opam and at
      https://github.com/ppedrot/ltac2.git branch v8.7)

    - UniCoq plugin (available at https://github.com/unicoq/unicoq.git
      branch v1.2-coq8.7-poly)

    - Mtac2 plugin (available at https://github.com/Mtac2/Mtac2.git
      branch master-8.7-poly)

    - Template-Coq plugin (available at
      https://github.com/Template-Coq/template-coq.git branch
      coq-8.7)

    The files which don't use these plugins will build fine without
    them. *)

(** This file defines common notations *)
Require Reify.Common.

(** This file contains the code used in the introduction *)
Require Reify.MiscIntro.

(** These files define PHOAS and some common wrapper tactics *)
Require Reify.PHOAS Reify.ReifyCommon.

(** Section 2.1 Typeclass-based reification *)
Require Reify.TypeClasses.
(** Version that uses the body rather than the type, but does not
    support [let ... in ...] *)
Require Reify.TypeClassesBodyFlatPHOAS.
(** Version that uses the body rather than the type, but reifies to
    [@expr nat] rather than [forall var, @expr var]. *)
Require Reify.TypeClassesBodyHOAS.

(** This file defines the primitive pair type used to track variables *)
Require Reify.PrimPair.

(** Section 2.2.Uncurrying *)
Require Reify.LtacPrimUncurry.

(** Section 2.2.Recursing under binders with typeclasses.Tracking with pairs *)
Require Reify.LtacTCPrimPair.

(** Section 2.2.Recursing under binders with typeclasses.Tracking with Gallina contexts *)
Require Reify.LtacTCGallinaCtx.

(** Section 2.2.Recursing under binders with typeclasses.Tracking with explicit contexts *)
Require Reify.LtacTCExplicitCtx.

(** Section 2.2.Recursing under binders with tactics in terms.Tracking with pairs *)
Require Reify.LtacTacInTermPrimPair.

(** Section 2.2.Recursing under binders with tactics in terms.Tracking with Gallina contexts *)
Require Reify.LtacTacInTermGallinaCtx.

(** Section 2.2.Recursing under binders with tactics in terms.Tracking with explicit contexts *)
Require Reify.LtacTacInTermExplicitCtx.

(** This file contains the scaffolding code to set up reification by canonical structures *)
Require Reify.CanonicalStructuresReifyCommon.

(** Section 2.3 Reification by canonical structures *)
(** The Flat* files take advantage of not needing to reify let-ins to
    avoid doing an initial [rewrite] to lock [Let_In].  The HOAS files
    reify to [@expr nat], which allows them to avoid some bookkeeping
    and be a bit shorter; the PHOAS files reify to [Expr]. *)
Require
  Reify.CanonicalStructuresFlatHOAS
  Reify.CanonicalStructuresFlatPHOAS
  Reify.CanonicalStructuresHOAS
  Reify.CanonicalStructuresPHOAS.

(** Section 2.4 Mtac *)
Require Reify.Mtac2.

(** Section 2.5 Ltac2 *)
(** This file contains the Ltac2 version of Ltac1 reification (Ltac1 version in LtacTacInTermExplicitCtx.v) *)
Require Reify.Ltac2.
(** This file contains low-level Ltac2 reification *)
Require Reify.Ltac2LowLevel.

(** Section 2.6 OCaml *)
(** The OCaml code itself is in reify_plugin.ml4 *)
Require Reify.OCaml.

(** Section 2.6 template-coq *)
Require Reify.TemplateCoq.

(** Section 2.6 quote plugin *)
Require Reify.QuoteFlat.

(** Section 3 Reification by parametricity *)
Require Reify.Parametricity.
(** This file contains the extra code to handle constants recursively *)
Require Reify.ParametricityWithConst.

(** This file defines some baseline statistics which are tested. *)
Require Reify.BaselineStats.

(** This file defines the terms being reified. *)
Require Reify.BenchmarkUtil.

(** The makefile target [parsing-test-files] will construct the .v
    files which reify by "copy-paste".  This target requires Bash and Python 2. *)

(** The makefile targets [quick-bench], [medium-bench], and
    [slow-bench] will run the benchmarks at various sizes. *)

(** The makefile target [bench.wl] will create a Wolfram Mathematica
    file with the relevant benchmark results in it, and the
    WolframScript file [reification-by-parametricity-graphs.wls] can
    use this file to generate the pdf graphs in the paper. *)

(** This file is a fully self-contained example of using reification by parametricity. *)
Require Reify.ExampleSelfContainedParametricity.
